/* dupsd  (Dumb UPS monitor Daemon)
 * Copyright (C) 1998 Malinin A. G.
 *
 * This  program  is free software;  you can redistribute it  and/or
 * modify  it  under  the  terms of the  GNU  General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed  in the hope that  it will be useful,
 * but  WITHOUT ANY WARRANTY;   without even the implied warranty of
 * MERCHANTABILITY  or  FITNESS  FOR  A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along  with  this  program;  if not,  write to the  Free Software
 * Foundation,  Inc.,   59  Temple  Place,  Suite  330,  Boston,  MA
 * 02111-1307  USA.
 *
 * Author's E-mail address: mag@h26.216.elnet.msk.ru.
 */

/* dups.c */


#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include <sys/ioctl.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include "msg.h"
#include "dups.h"

LINE_DESCR ld[shutdown_DUPS+1]={
  {TIOCM_CD, UNDEFINED},	/* state unused */
  {TIOCM_CTS, USUAL},		/* "on battery" state description */
  {TIOCM_RI, UNDEFINED},	/* "low battery" state description */
  {TIOCM_DSR, UNDEFINED},	/* "DUPS error" state description */
  {TIOCM_RTS, USUAL},		/* "monitor ready" command description */
  {TIOCM_DTR, USUAL}		/* "shutdown DUPS" command description */
};

/* delay for "Turn the UPS off" command */
unsigned int shutdown_delay=60;


/*
  opens the device "device", which should be a serial port
  that communicates with the UPS; returns a file descriptor;
  if open fails, writes a log message and returns -1
*/
int
open_DUPS(const char *device)
{
  int fd;

  if ((fd=open(device, O_RDWR|O_NDELAY|O_NOCTTY))==-1)
    log_msg(device, sys_errlist[errno], 1);
  return (fd);
}

/*
  returns logical EXCLUSIVE OR for logical arguments
*/
static int
xor(int arg1, int arg2)
{
  return ((!arg1&&arg2)||(arg1&&!arg2));
}

/*
  swithes the previously opened serial interface (with descriptor "fd";
  to which the UPS is attached) wire "line_descr_item"
  to logical level "1" (state==ON) or "0" (state==OFF)
*/
static void
switch_line(int fd, LINE_DESCR *line_descr_item, int state)
{
  int flags;

  switch (line_descr_item->logic) {
    case USUAL:
    case INVERSE:
      flags=line_descr_item->line;
      if (ioctl(fd, (xor(state, line_descr_item->logic))?TIOCMBIC:TIOCMBIS, &flags))
        log_msg("switch_line.ioctl", sys_errlist[errno], 1);
      break;
  }
}

/*
  returns current state of the previously opened serial interface
  (with descriptor "fd"; to which the UPS is attached) wire "line_descr_item"
  (ON|OFF|UNKNOWN)
*/
static int
check_line(int fd, int flags, LINE_DESCR *line_descr_item)
{

  switch (line_descr_item->logic) {
    case USUAL:
    case INVERSE:
      return (xor((flags&line_descr_item->line), line_descr_item->logic)?OFF:ON);
    default:
      return (UNKNOWN);
  }
}

/*
  initializes the UPS on previously opened device with descriptor "fd"
*/
void
setup_DUPS(int fd)
{
  switch_line(fd, &ld[shutdown_DUPS], OFF);
  switch_line(fd, &ld[monitor_ready], ON);
  sleep(1);
}

/*
  turns the UPS (on previously opened device with descriptor "fd") off
*/
void
kill_DUPS_power(int fd)
{
  log_msg("!", "killing the UPS power...\n", 1);
  switch_line(fd, &ld[shutdown_DUPS], ON);
  sleep(shutdown_delay);
}

/*
  returns the UPS (on previously opened device with descriptor "fd") state
  (DUPS_error|low_DUPS_battery|DUPS_on_battery|DUPS_OK)
*/
int
check_DUPS(int fd)
{
  int flags;

  if (ioctl(fd, TIOCMGET, &flags)) {
    log_msg("check_DUPS.ioctl", sys_errlist[errno], 1);
    return (DUPS_error);
  }
  else
    if (check_line(fd, flags, &ld[DUPS_error])==OFF)
      return (DUPS_error);
    else
      if (check_line(fd, flags, &ld[low_DUPS_battery])==ON)
        return (low_DUPS_battery);
      else
        if (check_line(fd, flags, &ld[DUPS_on_battery])==ON)
          return (DUPS_on_battery);
        else return (DUPS_OK);
}
