/* dupsd  (Dumb UPS monitor Daemon)
 * Copyright (C) 1998 Malinin A. G.
 *
 * This  program  is free software;  you can redistribute it  and/or
 * modify  it  under  the  terms of the  GNU  General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed  in the hope that  it will be useful,
 * but  WITHOUT ANY WARRANTY;   without even the implied warranty of
 * MERCHANTABILITY  or  FITNESS  FOR  A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along  with  this  program;  if not,  write to the  Free Software
 * Foundation,  Inc.,   59  Temple  Place,  Suite  330,  Boston,  MA
 * 02111-1307  USA.
 *
 * Author's E-mail address: mag@h26.216.elnet.msk.ru.
 */

/* dupsd.c */


#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>

#if HAVE_GETOPT_LONG
#  include <getopt.h>
#endif /* HAVE_GETOPT_LONG */

#include <signal.h>
#include <setjmp.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include "msg.h"
#include "dups.h"
#include "net_dups.h"
#include "version.h"

#ifndef INADDR_NONE
#  define INADDR_NONE 0xffffffff
#endif /* INADDR_NONE */

/* static variables which shadow parameters in the configuration file */
static int kill_power_mode=0;
static int server_mode=0;
static int client_mode=0;
static int batch_mode=0;
static char device[MAX_STR_LEN]="/dev/ttyS1";
static unsigned int poll_number=3;
static unsigned int poll_delay=5;
static int run_until_low_battery=0;
static unsigned long server=INADDR_NONE;
static int port=-10241;
static char power_state_file[MAX_STR_LEN]="/var/run/powerstatus";
static int _SIGPWR=SIGHUP;
static pid_t SIGPWR_handler_PID=1;
static char lock_file[MAX_STR_LEN]="/var/lock/dupsd/LCK..ttyS1";
static char log[MAX_STR_LEN]="/var/log/dupsd.log";
static int SIGTERM_received=0;
static jmp_buf jmp_env;	


/*
  notifies power state handling process about power state changes;
  if boolean parameter "power_on":
      >0 - write "OK" into the power state file
      =0 - write "FAIL" into the power state file*/
static void
notify_about_power_state(int power_on)
{
  if (*power_state_file) {
    int fd;

    unlink(power_state_file);
    if ((fd=open(power_state_file, O_CREAT|O_WRONLY|O_EXCL, 0644))!=-1) {
      char s[6];
      size_t l;

      l=strlen(strcpy(s, power_on?"OK\n":"FAIL\n"));
      if (write(fd, s, l)!=l)
        log_msg(power_state_file, "write error", 1);
      if (close(fd))
        log_msg(power_state_file, sys_errlist[errno], 1);
    }
    else log_msg(power_state_file, sys_errlist[errno], 1);
  }
  if (kill(SIGPWR_handler_PID, _SIGPWR))
    log_msg("notify_about_power_state.kill", sys_errlist[errno], 1);
}

/*
  verifies power state changes during time interval "poll_number*poll_delay";
  parameter "DUPS_state" contains the current UPS state ("on battery", etc)
*/
static void
verify_power_state(int DUPS_state)
{
  if ((DUPS_state>=unknown_DUPS_state)&&(DUPS_state<=DUPS_error)) {
    static int cnt=0;
    static int prev_DUPS_state=unknown_DUPS_state;

    if (prev_DUPS_state==DUPS_state)
      cnt=0;
    else
      if (cnt<poll_number)
        cnt++;
    if (cnt==poll_number) {
      log_msg("UPS state", (DUPS_state==unknown_DUPS_state)?"unknown state":state_str[DUPS_state], 1);
      cnt=0;
      switch (prev_DUPS_state=DUPS_state) {
        case low_DUPS_battery:
        case DUPS_error:
          notify_about_power_state(OFF);
          break;
        case DUPS_on_battery:
          if (!run_until_low_battery)
            notify_about_power_state(OFF);
          break;
        case DUPS_OK:
          notify_about_power_state(ON);
          break;
      }
    }
  }
  else log_msg("UPS state", "unrecognized state", 0);
}

/*
  converts internal description of the computer's serial
  interface wire "ld[line_item]" into text description "descr"
  for further configuration dumping; returns pointer to "descr"
*/
static char *
get_line(int line_item, char *descr, int len)
{
  if (descr&&(len>1)) {
    memset(descr, '\0', len);
    if (!((line_item<DUPS_OK)||(line_item>shutdown_DUPS))) {
      switch (ld[line_item].logic) {
        case INVERSE:
          *descr='~';
          len--;
        case USUAL:
          break;
        case UNDEFINED:
          return (descr);
        default:
          log_msg("get_line", "unrecognized line logic", 0);
          *descr='?';
          len--;
          break;
      }
      switch (ld[line_item].line) {
        case TIOCM_DTR:
          if (len>0)
            strncat(descr, "DTR", len-1);
          break;
        case TIOCM_RTS:
          if (len>0)
            strncat(descr, "RTS", len-1);
          break;
        case TIOCM_DSR:
          if (len>0)
            strncat(descr, "DSR", len-1);
          break;
        case TIOCM_CTS:
          if (len>0)
            strncat(descr, "CTS", len-1);
          break;
        case TIOCM_RI:
          if (len>0)
            strncat(descr, "RI", len-1);
          break;
        case TIOCM_CD:
          if (len>0)
            strncat(descr, "CD", len-1);
          break;
        default:
          log_msg("get_line", "unrecognized line logic", 0);
          if (len>0)
            strncat(descr, "?", len-1);
          break;
      }
    }
  }
  return (descr);
}

/*
  dumps current daemon configuration into the file "log"
*/
static void
dump_cfg(void)
{
  FILE *f;
  char li[5], s[MAX_STR_LEN];
  struct in_addr addr;

  if (*log) {
    if ((f=fopen(log, "a"))) {
      fprintf(f, "# Configuration parameters for %s\n\n", prg);
      sprintf(s, "device=%s\n", device);
      fprintf(f, "%s", s);
      sprintf(s, "monitor_ready=%s\n", get_line(monitor_ready, li, sizeof(li)));
      fprintf(f, "%s", s);
      sprintf(s, "DUPS_ready=%s\n", get_line(DUPS_error, li, sizeof(li)));
      fprintf(f, "%s", s);
      sprintf(s, "DUPS_on_battery=%s\n", get_line(DUPS_on_battery, li, sizeof(li)));
      fprintf(f, "%s", s);
      sprintf(s, "low_DUPS_battery=%s\n", get_line(low_DUPS_battery, li, sizeof(li)));
      fprintf(f, "%s", s);
      sprintf(s, "shutdown_DUPS=%s\n", get_line(shutdown_DUPS, li, sizeof(li)));
      fprintf(f, "%s", s);
      sprintf(s, "shutdown_delay=%u\n", shutdown_delay);
      fprintf(f, "%s", s);
      sprintf(s, "poll_number=%u\n", poll_number);
      fprintf(f, "%s", s);
      sprintf(s, "poll_delay=%u\n", poll_delay);
      fprintf(f, "%s", s);
      if (!run_until_low_battery)
        fprintf(f, "# ");
      fprintf(f, "run_until_low_battery\n");
      addr.s_addr=server;
      sprintf(s, "server=%s\n", client_mode?inet_ntoa(addr):"");
      fprintf(f, "%s", s);
      sprintf(s, "port=");
      fprintf(f, "%s", s);
      if (client_mode||server_mode)
        sprintf(s, "%d", port);
      else *s='\0';
      fprintf(f, "%s", strcat(s, "\n"));
      sprintf(s, "power_state_file=%s\n", power_state_file);
      fprintf(f, "%s", s);
      sprintf(s, "SIGPWR=%d\n", _SIGPWR);
      fprintf(f, "%s", s);
      sprintf(s, "SIGPWR_handler_PID=%lu\n", (unsigned long)SIGPWR_handler_PID);
      fprintf(f, "%s", s);
      sprintf(s, "lock_file=%s\n", lock_file);
      fprintf(f, "%s", s);
      sprintf(s, "log=%s\n", log);
      fprintf(f, "%s", s);
      if (ferror(f))
        log_msg("dump_cfg", sys_errlist[errno], 1);
      fclose(f);
    }
    else log_msg(log, sys_errlist[errno], 1);
  }
}

/*
  handles signals which dupsd receives
*/
static RETSIGTYPE
SIG_handler(int SIG)
{
  pid_t pid;

  signal(SIG, SIG_handler);
  switch (SIG) {
    case SIGCHLD:
      while ((pid=waitpid(-1, NULL, WNOHANG)))
        if ((pid==-1)&&(errno!=EINTR))
          break;
      break;
    case SIGTERM:
      SIGTERM_received=1;
      break;
    case SIGUSR1:
      longjmp(jmp_env, 0);
      break;
    case SIGUSR2:
      if (*log)
        dump_cfg();
      else log_msg("dumping configuration parameters", "undefined log file", 1);
      break;
  }
}

/*
  becomes a daemon
*/
static void
_daemon(void)
{
  if (!(kill_power_mode||batch_mode)) {
    long i, mfd;
    char s[MAX_STR_LEN];

    signal(SIGHUP, SIG_IGN);
    signal(SIGINT, SIG_IGN);
    signal(SIGQUIT, SIG_IGN);
    signal(SIGTSTP, SIG_IGN);
    signal(SIGTTOU, SIG_IGN);
    signal(SIGTTIN, SIG_IGN);
    signal(SIGPIPE, SIG_IGN);
    signal(SIGTERM, SIG_handler);
    signal(SIGCHLD, SIG_handler);
    signal(SIGUSR1, SIG_handler);
    signal(SIGUSR2, SIG_handler);
    if ((mfd=sysconf(_SC_OPEN_MAX))<=0) {
      sprintf(s, "%ld", mfd);
      log_msg("sysconf(_SC_OPEN_MAX)", s, 1);
      exit(1);
    }
    for (i=0; i<mfd; i++)
      close((int)i);
    switch (fork()) {
      case 0:
        setsid();
        chdir("/");
        umask(0);
        break;
      case -1:
        log_msg("_daemon.fork", sys_errlist[errno], 1);
        exit(1);
      default:
        exit(0);
    }
  }
}

/*
  converts server address "server" from name or dotted notation
  into internal representation and returns the last
*/
static unsigned long
get_server_addr(const char *server)
{
  unsigned long addr;
    
  if ((addr=inet_addr(server))==INADDR_NONE) {
    struct hostent *hep;

    if (!(hep=gethostbyname(server))) {
      log_msg("get_server_addr.gethostbyname", "error", 1);
      return (INADDR_NONE);
    }
    memcpy(&addr, hep->h_addr, hep->h_length);
  }
  return (addr);
}

/*
  initializes the computer's serial interface wire
  internal description "ld[line_item]" from description "descr"
  in the configuration file
*/
static void
set_line(int line_item, const char *descr)
{
  if (!((line_item<DUPS_OK)||(line_item>shutdown_DUPS))) {
    ld[line_item].line=TIOCM_CD;
    ld[line_item].logic=UNDEFINED;
    if (descr) {
      int flg=0;

      if (*descr=='~') {
        flg=1;
        descr++;
      }
      if (!strcmp(descr, "DTR"))
        ld[line_item].line=TIOCM_DTR;
      if (!strcmp(descr, "RTS"))
        ld[line_item].line=TIOCM_RTS;
      if (!strcmp(descr, "DSR"))
        ld[line_item].line=TIOCM_DSR;
      if (!strcmp(descr, "CTS"))
        ld[line_item].line=TIOCM_CTS;
      if (!strcmp(descr, "RI"))
        ld[line_item].line=TIOCM_RI;
      if (!strcmp(descr, "CD"))
        ld[line_item].line=TIOCM_CD;
      ld[line_item].logic=flg?INVERSE:USUAL;
    }
  }
}

/*
  writes message about dupsd usage into system log
*/
static void
usage(void)
{
  log_msg("COPYLEFT", prg, 1);
  log_msg("COPYLEFT", "Copyright (C) 1998 Malinun A. G.", 1);
  log_msg("COPYLEFT", "dupsd comes with ABSOLUTELY NO WARRANTY.  This is", 1);
  log_msg("COPYLEFT", "free software,  and you are welcome  to redistri-", 1);
  log_msg("COPYLEFT", "bute it under terms of GNU General Public License", 1);
  log_msg("COPYLEFT", "version 2.", 1);

#if HAVE_GETOPT_LONG
  log_msg("usage", "/sbin/dupsd [(-(b|s))|(--(check_UPS|shutdown_UPS))]", 1);
  log_msg("usage", "  [((-c)|(--conf)) <configuration_file>] ", 1);
#else /* not HAVE_GETOPT_LONG */
  log_msg("usage", "/sbin/dupsd [-(b|s)] [-c <configuration_file>] ", 1);
#endif /* HAVE_GETOPT_LONG */

  log_msg("usage", "  -b: check UPS and exit", 1);
  log_msg("usage", "  -s: shutdown UPS", 1);
  log_msg("usage", "  -c <configuration_file>: set configuration file", 1);

#if HAVE_GETOPT_LONG
  log_msg("usage", "  --check_UPS: check UPS and exit", 1);
  log_msg("usage", "  --shutdown_UPS: shutdown UPS", 1);
  log_msg("usage", "  --conf <configuration_file>: set configuration file", 1);
#endif /* HAVE_GETOPT_LONG */

}

/*
  reads command line and parameters from the configuration file
  and prepares internal structures and variables
*/
static void
read_parameters(int argc, char *argv[])
{
  FILE *cfg_f;
  char cfg_fn[MAX_STR_LEN]="/etc/DUPSd.cfg";
  int option;

#if HAVE_GETOPT_LONG
  int long_options_index=0;
  struct option long_options[]={
    {"check_UPS", 0, 0, 'b'},
    {"shutdown_UPS", 0, 0, 's'},
    {"conf", 1, 0, 'c'},
    {0, 0, 0, 0}
  };
#endif /* HAVE_GETOPT_LONG */

  kill_power_mode=client_mode=server_mode=batch_mode=0;
  shutdown_delay=poll_delay=run_until_low_battery=0;
  poll_number=1;
  _SIGPWR=SIGHUP;
  SIGPWR_handler_PID=1;
  server=INADDR_NONE;
  port=-10241;
  *device=*power_state_file=*lock_file=*log='\0';
  set_line(DUPS_OK, NULL);
  set_line(monitor_ready, NULL);
  set_line(DUPS_error, NULL);
  set_line(DUPS_on_battery, NULL);
  set_line(low_DUPS_battery, NULL);
  set_line(shutdown_DUPS, NULL);

  opterr=0;

#if HAVE_GETOPT_LONG
  while ((option=getopt_long(argc, argv, "bsc:", long_options, &long_options_index))!=EOF)
#else /* not HAVE_GETOPT_LONG */
  while ((option=getopt(argc, argv, "bsc:"))!=EOF)
#endif /* HAVE_GETOPT_LONG */

    switch (option) {
      case 'b':
        batch_mode=!(kill_power_mode=0);
        break;
      case 's':
        kill_power_mode=!(batch_mode=0);
        break;
      case 'c':
        if (optarg) {
          *cfg_fn='\0';
          strncat(cfg_fn, optarg, sizeof(cfg_fn)-1);
        }
        break;
      default:
        usage();
        exit(1);
    }
  if (optind<argc) {
    usage();
    exit(1);
  }
  if ((cfg_f=fopen(cfg_fn, "r"))) {
    unsigned long cnt=0;

    while (!feof(cfg_f)) {
      char s[MAX_STR_LEN+1], *pn;
      static const char *delimiters="=\n";

      pn=fgets(s, sizeof(s)-1, cfg_f);
      cnt++;
      if (!pn) {
        if (ferror(cfg_f)) {
          log_msg(cfg_fn, sys_errlist[errno], 1);
          exit(1);
        }
        else continue;
      }
      else
        if ((*s!='\0')&&(*s!='#'))
          if ((pn=(char *)strtok(s, delimiters))) {
            if (!strcmp(pn, "device")) {
              if ((pn=strtok(NULL, delimiters)))
                strcpy(device, pn);
              continue;
            }
            if (!strcmp(pn, "monitor_ready")) {
              if ((pn=strtok(NULL, delimiters)))
                set_line(monitor_ready, pn);
              continue;
            }
            if (!strcmp(pn, "DUPS_ready")) {
              if ((pn=strtok(NULL, delimiters)))
                set_line(DUPS_error, pn);
              continue;
            }
            if (!strcmp(pn, "DUPS_on_battery")) {
              if ((pn=strtok(NULL, delimiters)))
                set_line(DUPS_on_battery, pn);
              continue;
            }
            if (!strcmp(pn, "low_DUPS_battery")) {
              if ((pn=strtok(NULL, delimiters)))
                set_line(low_DUPS_battery, pn);
              continue;
            }
            if (!strcmp(pn, "shutdown_DUPS")) {
              if ((pn=strtok(NULL, delimiters)))
                set_line(shutdown_DUPS, pn);
              continue;
            }
            if (!strcmp(pn, "shutdown_delay")) {
              if ((pn=strtok(NULL, delimiters)))
                shutdown_delay=(unsigned int)strtoul(pn, NULL, 10);
              continue;
            }
            if (!strcmp(pn, "poll_number")) {
              if ((pn=strtok(NULL, delimiters)))
                if (!(poll_number=(unsigned int)strtoul(pn, NULL, 10)))
                  poll_number=1;
              continue;
            }
            if (!strcmp(pn, "poll_delay")) {
              if ((pn=strtok(NULL, delimiters)))
                poll_delay=(unsigned int)strtoul(pn, NULL, 10);
              continue;
            }
            if (!strcmp(pn, "run_until_low_battery")) {
              run_until_low_battery=1;
              continue;
            }
            if (!strcmp(pn, "server")) {
              client_mode=0;
              if ((pn=strtok(NULL, delimiters)))
                if (*pn) {
                  if ((server=get_server_addr(pn))==INADDR_NONE) {
                    log_msg(pn, "unrecognized server", 1);
                    fclose(cfg_f);
                    exit(1);
                  }
                  else client_mode=!(server_mode=0);
                }
              continue;
            }
            if (!strcmp(pn, "port")) {
              if ((pn=strtok(NULL, delimiters))) {
                if ((port=atoi(pn))<=0)
                  server_mode=client_mode=0;
                else server_mode=!client_mode;
              }
              continue;
            }
            if (!strcmp(pn, "SIGPWR")) {
              if ((pn=strtok(NULL, delimiters)))
                if ((_SIGPWR=atoi(pn))<=0)
                  _SIGPWR=SIGHUP;
              continue;
            }
            if (!strcmp(pn, "power_state_file")) {
              if ((pn=strtok(NULL, delimiters)))
                strcpy(power_state_file, pn);
              continue;
            }
            if (!strcmp(pn, "SIGPWR_handler_PID"))
              if ((pn=strtok(NULL, delimiters))) {
                if (!(kill_power_mode||batch_mode)) {
                  FILE *f;

                  if (*pn) {
                    if (!(f=fopen(pn, "r"))) {
                      log_msg(pn, sys_errlist[errno], 1);
                      fclose(cfg_f);
                      exit(1);
                    }
                    else {
                      char *sp;

                      sp=fgets(s, sizeof(s)-1, f);
                      if (!sp) {
                        if (ferror(cfg_f)) {
                          log_msg(pn, sys_errlist[errno], 1);
                          fclose(f);
                          fclose(cfg_f);
                          exit(1);
                        }
                      }
                      else SIGPWR_handler_PID=(pid_t)strtoul(s, NULL, 10);
                      fclose(f);
                      continue;
                    }
                  }
                }
                else continue;
                log_msg("SIGPWR_handler_PID", "unrecognized PID", 1);
                fclose(cfg_f);
                exit(1);
              }
            if (!strcmp(pn, "lock_file")) {
              if ((pn=strtok(NULL, delimiters)))
                strcpy(lock_file, pn);
              continue;
            }
            if (!strcmp(pn, "log")) {
              if ((pn=strtok(NULL, delimiters)))
                strcpy(log, pn);
              continue;
            }
            sprintf(s, "unrecognized line %lu", cnt);
            log_msg("reading configuration file", s, 1);
          }
    }
    fclose(cfg_f);
  }
  else {
    log_msg(cfg_fn, sys_errlist[errno], 1);
    exit(1);
  }
}

/*
  deletes "lock_file"
*/
static void
unlink_lock_file(void)
{
  if ((!(kill_power_mode||batch_mode))&&(*lock_file))
    if (unlink(lock_file))
      log_msg(lock_file, sys_errlist[errno], 1);
}

/*
  creates "lock_file"
*/
static void
create_lock_file(void)
{
  if ((!(kill_power_mode||batch_mode))&&(*lock_file)) {
    int fd;

    if ((fd=open(lock_file, O_CREAT|O_WRONLY|O_EXCL, 0644))==-1) {
      if (errno==EEXIST) {
        char s[3*MAX_STR_LEN];

        sprintf(s, "device %s locked by lock file %s", device, lock_file);
        log_msg("[re]starting dupsd", s, 1);
        exit(1);
      }
      else {
        log_msg(lock_file, sys_errlist[errno], 1);
        exit(1);
      }
    }
    else
      if (close(fd)) {
        log_msg(lock_file, sys_errlist[errno], 1);
        exit(1);
      }
  }
}

/*
  closes server socket with descriptor "server_fd"
  and the UPS device with descriptor "fd"
*/
static void
close_Fs(int fd)
{
  if (server_fd!=-1) {
    if (close(server_fd))
      log_msg("closing server socket", sys_errlist[errno], 0);
    server_fd=-1;
  }
  if (fd!=-1) {
    if (close(fd))
      log_msg(device, sys_errlist[errno], 1);
    fd=-1;
  }
}

int
main(int argc, char *argv[])
{
  static int current_DUPS_state=DUPS_OK;
  static int fd=-1;
  int longjmp_called;

  if (getuid()) {
    log_msg("!", "only root can start the process", 1);
    return (1);
  }
  *prg='\0';
  strncat(prg, package_name, sizeof(prg)-1);
  strncat(prg, " v. ", sizeof(prg)-1-strlen(prg));
  strncat(prg, package_version, sizeof(prg)-1-strlen(prg));
  if ((longjmp_called=setjmp(jmp_env))) {

    /* reread configuration */
    close_Fs(fd);
    unlink_lock_file();
    read_parameters(argc, argv);
    log_msg("configuration file", "read", 1);
  }
  else {

    /* read configuration and become a daemon */
    read_parameters(argc, argv);
    _daemon();
  }
  create_lock_file();
  if (!client_mode) {
    if ((fd=open_DUPS(device))==-1) {
      log_msg(device, sys_errlist[errno], 1);
      unlink_lock_file();
      return (1);
    }
    else setup_DUPS(fd);
  }
  {
    int r=0;

    /* turn the UPS off */
    if (kill_power_mode) {
      if (!client_mode)
        kill_DUPS_power(fd);
    }
    else
      if (batch_mode)

        /* return current UPS state and exit */
        r=100+(client_mode?check_net_DUPS(server, port):check_DUPS(fd));
      else {

        /* monitor the UPS state and serve clients */
        while (!SIGTERM_received) {
          if (server_mode)
            serve_net(port, poll_delay, current_DUPS_state);
          else sleep(poll_delay);
          verify_power_state(client_mode?check_net_DUPS(server, port):(current_DUPS_state=check_DUPS(fd)));
        }
        log_msg("stopping", "SIGTERM received", 1);
      }
    close_Fs(fd);
    unlink_lock_file();
    return (r);
  }
}
