/* dupsd  (Dumb UPS monitor Daemon)
 * Copyright (C) 1998 Malinin A. G.
 *
 * This  program  is free software;  you can redistribute it  and/or
 * modify  it  under  the  terms of the  GNU  General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed  in the hope that  it will be useful,
 * but  WITHOUT ANY WARRANTY;   without even the implied warranty of
 * MERCHANTABILITY  or  FITNESS  FOR  A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along  with  this  program;  if not,  write to the  Free Software
 * Foundation,  Inc.,   59  Temple  Place,  Suite  330,  Boston,  MA
 * 02111-1307  USA.
 *
 * Author's E-mail address: mag@h26.216.elnet.msk.ru.
 */

/* net_dups.c */


#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/time.h>
#include "msg.h"
#include "dups.h"

/* DUPS state (from DUPS_OK to DUPS_error) descriptions in plain text */
const char *state_str[DUPS_error+1]={
  "OK", "on battery", "low battery", "error"
};

/* master dupsd server socket descriptor */
int server_fd=-1;


/*
  opens the TCP connection with master dupsd, which runs at
  server "server" on port "port"; returns a socket descriptor;
  if open fails, writes a log message and returns -1
*/
static int
open_session(unsigned long server, int port)
{
  int fd;
    
  if ((fd=socket(AF_INET, SOCK_STREAM, 0))==-1)
    log_msg("open_session.socket", sys_errlist[errno], 1);
  else {
    struct sockaddr_in server_addr;

    memset(&server_addr, '\0', sizeof(server_addr));
    server_addr.sin_family=AF_INET;
    server_addr.sin_port=htons(port);
    server_addr.sin_addr.s_addr=server;
    if (connect(fd, (struct sockaddr *)&server_addr, sizeof(server_addr))==-1) {
      log_msg("open_session.connect", sys_errlist[errno], 1);
      close(fd);
      return (-1);
    }
    else log_msg("connecting with server", "OK", 0);
  }
  return (fd);
}

/*
  reads message "msg" with maximum length "msg_len"
  from TCP socket with descriptor "fd"; returns the number
  of read bytes
*/
static int
get_msg(int fd, char *msg, int msg_len)
{
  int read_cnt, left_cnt;

  left_cnt=msg_len;
  while (left_cnt>0) {
    if ((read_cnt=read(fd, msg, left_cnt))<0)		/* error */
      return (read_cnt);
    else
      if (!read_cnt)					/* EOF */
        break;
    left_cnt-=read_cnt;
    msg+=read_cnt;
  }
  return (msg_len-left_cnt);
}

/*
  writes message "msg" with maximum length "msg_len"
  into TCP socket with descriptor "fd"; returns the
  number of written bytes
*/
static int
put_msg(int fd, const char *msg, int msg_len)
{
  int write_cnt, left_cnt;

  left_cnt=msg_len;
  while (left_cnt>0) {
    if ((write_cnt=write(fd, msg, left_cnt))<=0)	/* error */
      return (write_cnt);
    left_cnt-=write_cnt;
    msg+=write_cnt;
  }
  return (msg_len-left_cnt);
}

/*
  returns current state (DUPS_OK, etc) of the network UPS
  at server "server" on port "port"
*/
int
check_net_DUPS(unsigned long server, int port)
{
  int i, fd;
  int DUPS_state=DUPS_error;
  char msg[MAX_STR_LEN];

  if ((fd=open_session(server, port))!=-1) {
    if ((i=get_msg(fd, msg, sizeof(msg)))>0) {
      if (close(fd))
        log_msg("closing connection with server", sys_errlist[errno], 1);
      else log_msg("closing connection with server", "OK", 0);
      for (i=DUPS_OK; i<=DUPS_error; i++)
        if (!strncmp(msg, state_str[i], sizeof(msg))) {
          DUPS_state=i;
          break;
        }
    }
    else
      if (i==-1)
        log_msg("check_net_DUPS.get_msg", sys_errlist[errno], 1);
      else
        if (!i)
          log_msg("check_net_DUPS.get_msg", "0 bytes received", 1);
  }
  return (DUPS_state);
}

/*
  master mode dupsd server: serves incoming TCP queries on port "port"
  from slave dupsd daemons during time interval "select_timeout"
  (reports to slave dupsd daemons "current_DUPS_state" (DUPS_OK, etc))
*/
void
serve_net(int port, int select_timeout, unsigned int current_DUPS_state)
{
  int client_fd;
  fd_set fds;
  int v;
  time_t start_time;
  struct timeval timeout;
  static struct sockaddr_in server_addr;
  static struct sockaddr_in client_addr;
  int client_addr_len;

  /* preparing server TCP socket for incoming connections */
  if (server_fd<0) {
    memset(&server_addr, '\0', sizeof(server_addr));
    server_addr.sin_family=AF_INET;
    server_addr.sin_port=htons(port);
    server_addr.sin_addr.s_addr=htonl(INADDR_ANY);
    if ((server_fd=socket(AF_INET, SOCK_STREAM, 0))==-1) {
      log_msg("serve_net.socket", sys_errlist[errno], 1);
      exit(1);
    }
    v=1;
    if (setsockopt(server_fd, SOL_SOCKET, SO_REUSEADDR, &v, sizeof(v))) {
      log_msg("serve_net.setsockopt(..., SO_REUSEADDR, ...)", sys_errlist[errno], 1);
      exit(1);
    }
    v=1;
    if (setsockopt(server_fd, SOL_SOCKET, SO_KEEPALIVE, &v, sizeof(v)))
      log_msg("serve_net.setsockopt(..., SO_KEEPALIVE, ...)", sys_errlist[errno], 1);
    v=1;
    if (setsockopt(server_fd, IPPROTO_TCP, TCP_NODELAY, &v, sizeof(v)))
      log_msg("serve_net.setsockopt(..., TCP_NODELAY, ...)", sys_errlist[errno], 1);
    if (bind(server_fd, (struct sockaddr *) &server_addr, sizeof(server_addr))) {
      log_msg("serve_net.bind", sys_errlist[errno], 1);
      exit(1);
    }
    if (listen(server_fd, 5)) {
      log_msg("serve_net.listen", sys_errlist[errno], 1);
      exit(1);
    }
  }

  /* waiting for TCP connections */
  if ((start_time=time(NULL))==(time_t)-1)
    log_msg("serve_net.time", "error", 0);
  while (1) {
    time_t end_time;

    FD_ZERO(&fds);
    FD_SET(server_fd, &fds);
    timeout.tv_sec=select_timeout;
    timeout.tv_usec=0;
    if ((v=select(server_fd+1, &fds, NULL, NULL, &timeout))==-1) {
      if (errno!=EINTR)
        log_msg("serve_net.select", sys_errlist[errno], 1);
    }
    else
      if (v>0) {

        /* serving incoming TCP connection by separate process */
        client_addr_len=sizeof(client_addr);
        if ((client_fd=accept(server_fd, (struct sockaddr *)&client_addr, &client_addr_len))==-1)
          log_msg("serve_net.accept", sys_errlist[errno], 1);
        else {
          char s[MAX_STR_LEN];
          struct in_addr ca;

          ca.s_addr=client_addr.sin_addr.s_addr;
          sprintf(s, "address=%s", inet_ntoa(ca));
          log_msg("client connected", s, 0);
          switch (fork()) {
            case -1:
              log_msg("serve_net.fork", sys_errlist[errno], 1);
              close(client_fd);
            case 0:
              if (close(server_fd))
                log_msg("serve_net.close(server_fd)", sys_errlist[errno], 1);
              if (put_msg(client_fd, state_str[current_DUPS_state], strlen(state_str[current_DUPS_state])+1)<=0)
                log_msg("serve_net.put_msg", sys_errlist[errno], 1);
              if (close(client_fd))
                log_msg(s, sys_errlist[errno], 1);
              else log_msg("client served", s, 0);
              exit(0);
            default:
              if (close(client_fd))
                log_msg("serve_net.close(client_fd)", sys_errlist[errno], 1);
          }
        }
      }

    /*
      if time interval is not over then check for next connection
      else return
    */
    if ((end_time=time(NULL))==(time_t)-1)
      log_msg("serve_net.time", "error", 0);
    else
      if (start_time!=(time_t)-1) {
        long interval;

        interval=end_time-start_time;
        if ((interval>=0)&&(interval<select_timeout)) {
          select_timeout-=interval;
          start_time=end_time;
          continue;
        }
      }
    return;
  }
}
