/* powerd (Power Daemon). Copyright (C) 1998 Malinin A. G.
 *
 * This  program  is free software;  you can redistribute it  and/or
 * modify  it  under  the  terms of the  GNU  General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed  in the hope that  it will be useful,
 * but  WITHOUT ANY WARRANTY;   without even the implied warranty of
 * MERCHANTABILITY  or  FITNESS  FOR  A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along  with  this  program;  if not,  write to the  Free Software
 * Foundation,  Inc.,   59  Temple  Place,  Suite  330,  Boston,  MA
 * 02111-1307  USA.
 *
 * Author's E-mail address: mag@h26.216.elnet.msk.ru.
 */

/* powerd.c */


#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#if HAVE_GETOPT_LONG
#  include <getopt.h>
#endif /* HAVE_GETOPT_LONG */

#include <signal.h>
#include <setjmp.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include "msg.h"
#include "version.h"

/* static variables which shadow parameters in the configuration file */
static char power_state_file[MAX_STR_LEN]="/var/run/powerstatus";
static char PID[MAX_STR_LEN]="/var/run/powerd.pid";
static char power_fail_script[MAX_STR_LEN]="/etc/rc.power start";
static char power_ok_script[MAX_STR_LEN]="/etc/rc.power stop";
static char log[MAX_STR_LEN]="/var/log/powerd.log";
static int SIGTERM_received=0;
static jmp_buf jmp_env;	


/*
  executes appropriate external command which depends on
  boolean parameter "power_fail":
      power_fail!=0 - "power_fail_script"
      power_fail==0 - "power_OK_script"
*/
static void
rc(int power_fail)
{
  int r;
  char s[MAX_STR_LEN], *rcp;

  switch (fork()) {
    case 0:
      if (system(NULL))
        switch (r=system(rcp=power_fail?power_fail_script:power_ok_script)) {
          case -1:
            log_msg("rc.system", sys_errlist[errno], 1);
            break;
          case 127:
            log_msg("rc.system", "shell exec error", 1);
            break;
          default:
            sprintf(s, "return code for %s is %d", rcp, r);
            log_msg("rc.system", s, 0);
        }
      else log_msg("rc.system", "shell is not available", 1);
      exit(0);
    case -1:
      log_msg("rc.fork", sys_errlist[errno], 1);
  }
}

/*
  checks current power state and executes the appropriate external command
*/
static void
check_power_state(void)
{
  if (*power_state_file) {
    FILE *f;
    int r=0;

    if ((f=fopen(power_state_file, "r"))) {
      char s[5+1], *pn;

      if ((pn=fgets(s, sizeof(s)-1, f))) {
        if (strstr(s, "OK")) {
          rc(0);
          r=1;
        }
        else
          if (strstr(s, "FAIL")) {
            rc(1);
            r=1;
          }
      }
      else
        if (ferror(f))
          log_msg(power_state_file, sys_errlist[errno], 1);
      if (!r)
        log_msg(power_state_file, "unrecognized contents", 1);
      if (fclose(f))
        log_msg(power_state_file, sys_errlist[errno], 1);
    }
    else log_msg(power_state_file, sys_errlist[errno], 1);
  }
  else log_msg("check_power_state", "undefined power state file", 1);
}

/*
  dumps current daemon configuration into the "log" file
*/
static void
dump_cfg(void)
{
  FILE *f;
  char s[MAX_STR_LEN];

  if (*log) {
    if ((f=fopen(log, "a"))) {
      fprintf(f, "# Configuration parameters for %s\n\n", prg);
      sprintf(s, "power_state_file=%s\n", power_state_file);
      fprintf(f, "%s", s);
      sprintf(s, "PID=%s\n", PID);
      fprintf(f, "%s", s);
      sprintf(s, "power_fail_script=%s\n", power_fail_script);
      fprintf(f, "%s", s);
      sprintf(s, "power_ok_script=%s\n", power_ok_script);
      fprintf(f, "%s", s);
      sprintf(s, "log=%s\n", log);
      fprintf(f, "%s", s);
      if (ferror(f))
        log_msg("dump_cfg", sys_errlist[errno], 1);
      fclose(f);
    }
    else log_msg(log, sys_errlist[errno], 1);
  }
}

/*
  handles signals which powerd receives
*/
static RETSIGTYPE
SIG_handler(int SIG)
{
  pid_t pid;

  signal(SIG, SIG_handler);
  switch (SIG) {
    case SIGCHLD:
      while ((pid=waitpid(-1, NULL, WNOHANG)))
        if ((pid==-1)&&(errno!=EINTR))
          break;
      break;
    case SIGTERM:
      SIGTERM_received=1;
      break;
    case SIGHUP:
      check_power_state();
      break;
    case SIGUSR1:
      longjmp(jmp_env, 0);
      break;
    case SIGUSR2:
      if (*log)
        dump_cfg();
      else log_msg("dumping configuration parameters", "undefined log file", 1);
      break;
  }
}

/*
  becomes a daemon
*/
static void
_daemon(void)
{
  long i, mfd;
  char s[MAX_STR_LEN];

  signal(SIGINT, SIG_IGN);
  signal(SIGQUIT, SIG_IGN);
  signal(SIGTSTP, SIG_IGN);
  signal(SIGTTOU, SIG_IGN);
  signal(SIGTTIN, SIG_IGN);
  signal(SIGHUP, SIG_handler);
  signal(SIGTERM, SIG_handler);
  signal(SIGCHLD, SIG_handler);
  signal(SIGUSR1, SIG_handler);
  signal(SIGUSR2, SIG_handler);
  if ((mfd=sysconf(_SC_OPEN_MAX))<=0) {
    sprintf(s, "%ld", mfd);
    log_msg("sysconf(_SC_OPEN_MAX)", s, 1);
    exit(1);
  }
  for (i=0; i<mfd; i++)
    close((int)i);
  switch (fork()) {
    case 0:
      setsid();
      chdir("/");
      umask(0);
      break;
    case -1:
      log_msg("_daemon.fork", sys_errlist[errno], 1);
      exit(1);
    default:
      exit(0);
  }
}

/*
  writes message about powerd usage into system log
*/
static void
usage(void)
{
  log_msg("COPYLEFT", prg, 1);
  log_msg("COPYLEFT", "Copyright (C) 1998 Malinun A. G.", 1);
  log_msg("COPYLEFT", "powerd comes with ABSOLUTELY NO WARRANTY. This is", 1);
  log_msg("COPYLEFT", "free software,  and you are welcome  to redistri-", 1);
  log_msg("COPYLEFT", "bute it under terms of GNU General Public License", 1);
  log_msg("COPYLEFT", "version 2.", 1);

#if HAVE_GETOPT_LONG
  log_msg("usage", "powerd [(-c)|(--conf) <configuration_file>]", 1);
#else /* not HAVE_GETOPT_LONG */
  log_msg("usage", "powerd [-c <configuration_file>]", 1);
#endif /* HAVE_GETOPT_LONG */

  log_msg("usage", "  -c <configuration_file>: set configuration file", 1);

#if HAVE_GETOPT_LONG
  log_msg("usage", "  --conf <configuration_file>: set configuration file", 1);
#endif /* HAVE_GETOPT_LONG */
}

/*
  reads command line and parameters from the configuration file
  and prepares internal structures and variables
*/
static void
read_parameters(int argc, char *argv[])
{
  FILE *cfg_f;
  char cfg_fn[MAX_STR_LEN]="/etc/powerd.cfg";
  int option;

#if HAVE_GETOPT_LONG
  int long_options_index=0;
  struct option long_options[]={ {"conf", 1, 0, 'c'}, {0, 0, 0, 0} };
#endif /* HAVE_GETOPT_LONG */

  *PID=*power_state_file=*power_fail_script=*power_ok_script=*log='\0';

  opterr=0;

#if HAVE_GETOPT_LONG
  while ((option=getopt_long(argc, argv, "c:", long_options, &long_options_index))!=EOF)
#else /* not HAVE_GETOPT_LONG */
  while ((option=getopt(argc, argv, "c:"))!=EOF)
#endif /* HAVE_GETOPT_LONG */

    switch (option) {
      default:
        usage();
        exit(1);
      case 'c':
        if (optarg) {
          *cfg_fn='\0';
          strncat(cfg_fn, optarg, sizeof(cfg_fn)-1);
        }
    }
  if (optind<argc) {
    usage();
    exit(1);
  }
  if ((cfg_f=fopen(cfg_fn, "r"))) {
    unsigned long cnt=0;

    while (!feof(cfg_f)) {
      char s[MAX_STR_LEN+1], *pn;
      static const char *delimiters="=\n";

      pn=fgets(s, sizeof(s)-1, cfg_f);
      cnt++;
      if (!pn) {
        if (ferror(cfg_f)) {
          log_msg(cfg_fn, sys_errlist[errno], 1);
          exit(1);
        }
        else continue;
      }
      else
        if ((*s!='\0')&&(*s!='#'))
          if ((pn=(char *)strtok(s, delimiters))) {
            if (!strcmp(pn, "power_state_file")) {
              if ((pn=strtok(NULL, delimiters)))
                strcpy(power_state_file, pn);
              continue;
            }
            if (!strcmp(pn, "PID")) {
              if ((pn=strtok(NULL, delimiters)))
                strcpy(PID, pn);
              continue;
            }
            if (!strcmp(pn, "power_fail_script")) {
              if ((pn=strtok(NULL, delimiters)))
                strcpy(power_fail_script, pn);
              continue;
            }
            if (!strcmp(pn, "power_ok_script")) {
              if ((pn=strtok(NULL, delimiters)))
                strcpy(power_ok_script, pn);
              continue;
            }
            if (!strcmp(pn, "log")) {
              if ((pn=strtok(NULL, delimiters)))
                strcpy(log, pn);
              continue;
            }
            sprintf(s, "unrecognized line %lu", cnt);
            log_msg("reading configuration file", s, 1);
          }
    }
    fclose(cfg_f);
  }
  else {
    log_msg(cfg_fn, sys_errlist[errno], 1);
    exit(1);
  }
}

/*
  deletes "PID" file
*/
static void
unlink_PID_file(void)
{
  if (*PID)
    if (unlink(PID))
      log_msg(PID, sys_errlist[errno], 1);
}

/*
  creates "PID" file
*/
static void
create_PID_file(void)
{
  if (*PID) {
    int fd;

    if ((fd=open(PID, O_CREAT|O_WRONLY|O_EXCL, 0644))==-1) {
      if (errno==EEXIST) {
        char s[2*MAX_STR_LEN];

        sprintf(s, "PID file %s exists", PID);
        log_msg("[re]starting powerd", s, 1);
        exit(1);
      }
      else {
        log_msg(PID, sys_errlist[errno], 1);
        exit(1);
      }
    }
    else {
      char s[MAX_STR_LEN];
      size_t l;

      sprintf(s, "%lu", (unsigned long)getpid());
      l=strlen(s);
      if (write(fd, s, l)!=l) {
        log_msg(PID, "write error", 1);
        exit(1);
      }
      if (close(fd)) {
        log_msg(PID, sys_errlist[errno], 1);
        exit(1);
      }
    }
  }
}

int
main(int argc, char *argv[])
{
  int longjmp_called;

  if (getuid()) {
    log_msg("!", "only root can start the process", 1);
    return (1);
  }
  *prg='\0';
  strncat(prg, "powerd v. ", sizeof(prg)-1);
  strncat(prg, package_version, sizeof(prg)-1-strlen(prg));
  if ((longjmp_called=setjmp(jmp_env))) {

    /* reread configuration parameters */
    read_parameters(argc, argv);
    log_msg("configuration file", "read", 1);
  }
  else {

    /* read configuration parameters and become a daemon */
    read_parameters(argc, argv);
    _daemon();
    create_PID_file();
  }

  /* handle incoming signals */
  while (!SIGTERM_received)
    sleep(1);
  log_msg("stopping", "SIGTERM received", 1);
  unlink_PID_file();
  return (0);
}
